RWTexture2D<float4> tResult : register(u0);

cbuffer SceneConstants : register( b0 )
{
    float4x4 mView;
    float4x4 mProjection;
    float4x4 mViewProjection;
    float4x4 mInvViewProjection;
    float4 vViewDir;
    float4 vViewPos;
    float4 deltas;
    float4 padd_[2];
};

Texture2D<float4> image : register(t0);
Texture2D<float4> gbNormals : register(t1);
Texture2D<float4> gbPos : register(t2);

SamplerState sampl_st : register(s0);

float gaussR(float sigma, float dist)
{
	return exp(-(dist*dist)/(2.0*sigma*sigma));
}

float  gaussD(float sigma, int x, int y)
{
	return exp(-((x*x+y*y)/(2.0f*sigma*sigma)));
}

#define MINF asfloat(0xff800000)

[numthreads(8, 8, 1)]
void Blur(int3 DTid : SV_DispatchThreadID, uint groupid : SV_GroupIndex)
{
    int screen_width = 1600;
    int screen_height = 900;
    int gb_screen_width = 1600;
    int gb_screen_height = 900;
	float sigmaD=0.5f;
	float sigmaR=1.5f;
    int2 posId = int2((float(DTid.x)/float(gb_screen_width))*screen_width,
                      (float(DTid.y)/float(gb_screen_height))*screen_height);
    float4 intCenter = image[posId.xy];

    float4 posCenter = gbPos[DTid.xy];
    float4 normCenter = gbNormals[DTid.xy];
    int kernelRadius = 0;
    float3 sum = intCenter.rgb;
	float sumWeight = 1.0f;
	

    for(int m = DTid.x-kernelRadius; m <= DTid.x+kernelRadius; m++)
    {
        for(int n = DTid.y-kernelRadius; n <= DTid.y+kernelRadius; n++)
        {
			if(m == DTid.x || n == DTid.y)
				continue;
            uint2 pos = uint2(m, n);
            int2 posId_2 = int2((float(pos.x)/float(gb_screen_width))*screen_width,
                                (float(pos.y)/float(gb_screen_height))*screen_height);
            float4 intKerPos = image[posId_2];
            
            float4 intKerNorm = gbNormals[pos];
            float d = distance(intKerPos.xyz, intCenter.xyz);
			float3 light_dir = gbPos[DTid.xy].xyz - float3(100,100,100);
            float3 view_dir = gbPos[DTid.xy].xyz - vViewPos.xyz;
			float depth_diff = distance(gbPos[DTid.xy].xyz, vViewPos.xyz);
			float depth_diff_2 = distance(vViewPos.xyz,posCenter.xyz);
			float normalCloseness = dot(intKerNorm.xyz, normCenter.xyz);
			normalCloseness = normalCloseness*normalCloseness;
			normalCloseness = normalCloseness*normalCloseness;
			#define DEPTH_WEIGHT 1.3f
			#define NORMAL_WEIGHT 1.0f
			float normalError = (1.0 - normalCloseness);
			float normalWeight = max((1.0 - normalError * NORMAL_WEIGHT), 0.00);
			float depthWeight = max(0.0, 1.0 - abs(depth_diff - depth_diff_2) * DEPTH_WEIGHT);
			
            float weight = depthWeight*normalWeight;
            if(weight > 0.0)
            {
                sumWeight += weight;
                sum += intKerPos.rgb*weight;
            }
        }
    }
    tResult[DTid.xy] = float4((sum/sumWeight), 1);
}